﻿namespace TabsStudioSorter
{
    class Comparer
    {
        public Comparer(Options options_, TabsStudioExt.Util.IsPreviewTabProperty isPreviewTabProperty_,
            System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo_,
            Marker marker_)
        {
            options = options_;
            isPreviewTabProperty = isPreviewTabProperty_;
            fixedInfo = fixedInfo_;
            marker = marker_;
        }

        public int CompareTabs(System.Windows.Controls.TabItem lhs, System.Windows.Controls.TabItem rhs)
        {
            if (isPreviewTabProperty.IsPreviewTab(lhs))
                return 1;
            if (isPreviewTabProperty.IsPreviewTab(rhs))
                return -1;
            if (!fixedInfo.ContainsKey(lhs))
                return 1;
            if (!fixedInfo.ContainsKey(rhs))
                return -1;

            FixedTabInfo lhsInfo = fixedInfo[lhs];
            FixedTabInfo rhsInfo = fixedInfo[rhs];
            foreach (SortingRule rule in options.rules)
            {
                int result = CompareByRule(lhs, rhs, lhsInfo, rhsInfo, rule);
                if (result != 0)
                    return result;
            }
            return 0;
        }

        private int CompareByRule(System.Windows.Controls.TabItem lhs, System.Windows.Controls.TabItem rhs,
            FixedTabInfo lhsInfo, FixedTabInfo rhsInfo,
            SortingRule rule)
        {
            int result = 0;
            switch (rule.type)
            {
                case SortingRuleType.MarkerHighlightIndex:
                    uint lhsIndex = marker.TabHighlightIndex(lhs);
                    uint rhsIndex = marker.TabHighlightIndex(rhs);
                    if (lhsIndex == 0)
                        lhsIndex = uint.MaxValue;
                    if (rhsIndex == 0)
                        rhsIndex = uint.MaxValue;
                    if (lhsIndex != rhsIndex)
                    {
                        return ApplyOrderRule(lhsIndex < rhsIndex ? -1 : 1, rule);
                    }
                    break;
                case SortingRuleType.Project:
                    result = CompareMatchedStrings(lhsInfo.project, rhsInfo.project, rule);
                    if (result != 0)
                        return ApplyOrderRule(result, rule);
                    break;
                case SortingRuleType.Extension:
                    result = CompareMatchedStrings(lhsInfo.extension, rhsInfo.extension, rule);
                    if (result != 0)
                        return ApplyOrderRule(result, rule);
                    break;
                case SortingRuleType.TabName:
                    result = CompareMatchedStrings(RuleMatch.GetTabNameForSorting(lhs, false), RuleMatch.GetTabNameForSorting(rhs, false), rule);
                    if (result != 0)
                        return ApplyOrderRule(result, rule);
                    break;
                case SortingRuleType.TabNameWithPath:
                    result = CompareMatchedStrings(RuleMatch.GetTabNameForSorting(lhs, true), RuleMatch.GetTabNameForSorting(rhs, true), rule);
                    if (result != 0)
                        return ApplyOrderRule(result, rule);
                    break;
                case SortingRuleType.IsDocument:
                    if (lhsInfo.isDocument != rhsInfo.isDocument)
                        return ApplyOrderRule(lhsInfo.isDocument ? -1 : 1, rule);
                    break;
                case SortingRuleType.Path:
                    result = CompareMatchedStrings(lhsInfo.paths, rhsInfo.paths, rule);
                    if (result != 0)
                        return ApplyOrderRule(result, rule);
                    break;
            }
            return 0;
        }

        private int CompareMatchedStrings(string lhs, string rhs, SortingRule rule)
        {
            string lhsMatch = RuleMatch.ExtractM(lhs, rule);
            string rhsMatch = RuleMatch.ExtractM(rhs, rule);

            if (lhsMatch == null && rhsMatch == null)
                return 0;

            if (lhsMatch != null && rhsMatch != null)
                return System.String.Compare(lhsMatch, rhsMatch, true);

            return lhsMatch != null ? -1 : 1;
        }

        private int ApplyOrderRule(int result, SortingRule rule)
        {
            return rule.inverseOrder ? -result : result;
        }

        private Options options;
        private TabsStudioExt.Util.IsPreviewTabProperty isPreviewTabProperty;
        private System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo;
        private Marker marker;
    }
}
