﻿namespace TabsStudioSorter
{
    public class Sorter : TabsStudioExt.ITabsStudioAddin, TabsStudioExt.IConfigurable
    {
        public Sorter()
        {
            resortMenuItem = new System.Windows.Controls.MenuItem();
            resortMenuItem.Header = "Re-sort tabs";
            resortMenuItem.Click += OnResort;
        }

        public void OnConnection(TabsStudioExt.ITabsStudioEngine engine_, EnvDTE80.DTE2 dte)
        {
            engine = engine_;
            storage = new OptionsStorage(engine.GetAddinSettingsDirectory());
            options = storage.Load();
            engine.TabCreated += OnTabCreated;
            engine.TabDestroyed += OnTabDestroyed;
            engine.TabRenamed += OnTabRenamed;
            engine.OpeningTabExtensionContextMenu += OpeningTabExtensionContextMenu;
            engine.OpeningTabContextMenu += OpeningTabContextMenu;
            priorityGroup = new PriorityGroup();
            priorityGroup.OnTopChanged += OnTopChanged;
            marker = new Marker();
            marker.OnHighlightIndexChanged += OnHighlightIndexChanged;
            fixedInfo = new System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo>();
            isPreviewTabProperty = new TabsStudioExt.Util.IsPreviewTabProperty();
            pathsProperty = new TabsStudioExt.Util.PathsProperty();
            comparer = new Comparer(options, isPreviewTabProperty, fixedInfo, marker);
            superGroup = new SuperGroup(options, priorityGroup, marker, isPreviewTabProperty, fixedInfo);
        }

        public void OnDisconnection()
        {
            engine.TabCreated -= OnTabCreated;
            engine.TabDestroyed -= OnTabDestroyed;
            engine.TabRenamed -= OnTabRenamed;
            engine.OpeningTabExtensionContextMenu -= OpeningTabExtensionContextMenu;
            engine.OpeningTabContextMenu -= OpeningTabContextMenu;
            priorityGroup.OnTopChanged -= OnTopChanged;
            marker.OnHighlightIndexChanged -= OnHighlightIndexChanged;
        }

        public void Configure(System.Windows.Window parent)
        {
            OptionsDialog dialog = new OptionsDialog(options);
            dialog.Owner = parent;
            dialog.WindowStartupLocation = System.Windows.WindowStartupLocation.CenterOwner;
            if (dialog.ShowDialog().GetValueOrDefault())
            {
                foreach (TabsStudioExt.Tabs tabs in engine.GetTabs())
                    Resort(tabs);
                storage.Save(options);
            }
        }

        private void OnTopChanged(System.Windows.Controls.TabItem tabItem, System.Windows.Controls.Panel panel)
        {
            SortTab(tabItem, panel);
        }

        private void OnHighlightIndexChanged(System.Windows.Controls.TabItem tabItem, System.Windows.Controls.Panel panel)
        {
            SortTab(tabItem, panel);
        }

        private void OnTabCreated(object sender, TabsStudioExt.TabEventArgs e)
        {
            fixedInfo[e.Tab.TabItem] = GetFixedTabInfo(e.Tab);
            SortTab(e.Tab.TabItem, e.Tabs.TabPanel2);
            priorityGroup.OnTabCreated(e);
            marker.OnTabCreated(e);
        }

        private void OnTabDestroyed(object sender, TabsStudioExt.TabEventArgs e)
        {
            superGroup.UpdateDeletedProperties(e.Tab.TabItem, e.Tabs.TabPanel2);
            fixedInfo.Remove(e.Tab.TabItem);
            priorityGroup.OnTabDestroyed(e);
            marker.OnTabDestroyed(e);
        }

        private void OnTabRenamed(object sender, TabsStudioExt.TabEventArgs e)
        {
            SortTab(e.Tab.TabItem, e.Tabs.TabPanel2);
        }

        private void OpeningTabExtensionContextMenu(object sender, TabsStudioExt.OpeningTabExtensionContextMenuEventArgs e)
        {
            storedTabs = e.Tabs;
            PlaceMenuItem(e.ContextMenu);
        }

        private void OpeningTabContextMenu(object sender, TabsStudioExt.OpeningTabContextMenuEventArgs e)
        {
            storedTabs = e.Tabs;
            PlaceMenuItem(e.ContextMenu);
        }

        private void Resort(TabsStudioExt.Tabs tabs)
        {
            foreach (TabsStudioExt.Tab tab in tabs.TabList)
                fixedInfo[tab.TabItem] = GetFixedTabInfo(tab);

            System.Collections.Generic.List<System.Windows.Controls.TabItem> items =
                new System.Collections.Generic.List<System.Windows.Controls.TabItem>();
            foreach (System.Windows.Controls.TabItem tabItem in tabs.TabPanel2.Children)
                items.Add(tabItem);

            foreach (System.Windows.Controls.TabItem tabItem in items)
                SortTab(tabItem, tabs.TabPanel2);
        }

        private void OnResort(object sender, System.Windows.RoutedEventArgs e)
        {
            Resort(storedTabs);
        }

        private void PlaceMenuItem(System.Windows.Controls.ContextMenu menu)
        {
            foreach (object item in menu.Items)
            {
                if (item is System.Windows.Controls.Control)
                {
                    if ((item as System.Windows.Controls.Control).Name.Equals("TabsStudioSeparator"))
                    {
                        menu.Items.Insert(menu.Items.IndexOf(item), new System.Windows.Controls.Separator());
                        menu.Items.Insert(menu.Items.IndexOf(item), resortMenuItem);
                        break;
                    }
                }
            }
        }

        private void SortTab(System.Windows.Controls.TabItem tabItem, System.Windows.Controls.Panel panel)
        {
            Range range = PrioritySortTab(tabItem, panel);
            bool priority = priorityGroup.IsPriorityTab(tabItem);
            for (int i = range.begin; i < range.end; ++i)
            {
                System.Windows.Controls.TabItem currentTabItem = (System.Windows.Controls.TabItem)panel.Children[i];
                if (currentTabItem != tabItem && priority == priorityGroup.IsPriorityTab(currentTabItem))
                {
                    if (panel.Children.IndexOf(currentTabItem) < panel.Children.IndexOf(tabItem) &&
                        comparer.CompareTabs(currentTabItem, tabItem) > 0)
                    {
                        panel.Children.Remove(tabItem);
                        panel.Children.Insert(panel.Children.IndexOf(currentTabItem), tabItem);
                        break;
                    }
                    if (panel.Children.IndexOf(currentTabItem) > panel.Children.IndexOf(tabItem) &&
                        comparer.CompareTabs(currentTabItem, tabItem) < 0)
                    {
                        panel.Children.Remove(tabItem);
                        panel.Children.Insert(panel.Children.IndexOf(currentTabItem) + 1, tabItem);
                    }
                }
            }
            superGroup.UpdateProperties(panel);
        }

        private Range PrioritySortTab(System.Windows.Controls.TabItem tabItem, System.Windows.Controls.Panel panel)
        {
            bool isPriority = priorityGroup.IsPriorityTab(tabItem);
            if (isPriority)
            {
                for (int i = 0; i < panel.Children.IndexOf(tabItem); ++i)
                {
                    if (!priorityGroup.IsPriorityTab((System.Windows.Controls.TabItem)panel.Children[i]))
                    {
                        panel.Children.Remove(tabItem);
                        panel.Children.Insert(i, tabItem);
                        return new Range(0, i + 1);
                    }
                }
            }
            else
            {
                for (int i = panel.Children.Count - 1; i > panel.Children.IndexOf(tabItem); --i)
                {
                    if (priorityGroup.IsPriorityTab((System.Windows.Controls.TabItem)panel.Children[i]))
                    {
                        panel.Children.Remove(tabItem);
                        panel.Children.Insert(i, tabItem);
                        return new Range(i, panel.Children.Count);
                    }
                }
            }
            return new Range(0, panel.Children.Count);
        }

        private FixedTabInfo GetFixedTabInfo(TabsStudioExt.Tab tab)
        {
            EnvDTE.Window w = tab.TabExtensions[0].Window;

            string project = null;
            {
                try
                {
                    if (w.Document != null && w.Document.ProjectItem != null &&
                        w.Document.ProjectItem.ContainingProject != null)
                    {
                        project = w.Document.ProjectItem.ContainingProject.Name;
                    }
                }
                catch (System.Exception)
                {
                }
            }

            string extension = null;
            {
                try
                {
                    if (w.Document != null)
                        extension = tab.TabExtensions[0].TitleManager.Title;
                }
                catch (System.Exception)
                {
                }
            }

            bool isDocument = false;
            {
                try
                {
                    isDocument = w.Document != null;
                }
                catch (System.Exception)
                {
                }
            }

            return new FixedTabInfo(project, extension, isDocument, pathsProperty.Paths(tab.TabItem));
        }

        private OptionsStorage storage;
        private Options options;
        private TabsStudioExt.ITabsStudioEngine engine;
        private TabsStudioExt.Tabs storedTabs;
        private System.Windows.Controls.MenuItem resortMenuItem;
        private PriorityGroup priorityGroup;
        private Marker marker;
        private System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo;
        private TabsStudioExt.Util.IsPreviewTabProperty isPreviewTabProperty;
        private TabsStudioExt.Util.PathsProperty pathsProperty;
        private Comparer comparer;
        private SuperGroup superGroup;
    }
}
